# @author Akash Pallath
# This code is licensed under the MIT license (see LICENSE.txt for details).
"""
Content negotiators to fetch citations in different formats.
"""
from collections import OrderedDict
import logging

import rispy

from paperfetcher.apiclients import CrossrefQuery
from paperfetcher.exceptions import ContentNegotiationError, RISParsingError

# Logging
logger = logging.getLogger(__name__)


def crossref_negotiate_ris(doi):
    """Fetches citation corresponding to a DOI in RIS format by using the
    Crossref REST API.

    Args:
        doi: DOI to fetch citation for.

    Returns:
        dict: rispy-read dictionary of RIS content returned by Crossref REST API.
    """
    # Build content negotation query
    components = OrderedDict([("works", doi),
                              ("transform", "application/x-research-info-systems")])
    query = CrossrefQuery(components)

    # Execute query
    query()

    if not (query.response.status_code == 200):
        raise ContentNegotiationError("Could not get RIS metadata for DOI %s" % doi)

    # rispy conversion
    try:
        ris_data = rispy.loads(query.response.text)
        return ris_data

    except Exception:
        raise RISParsingError("Could not parse RIS metadata returned by Crossref for DOI %s" % doi)



# @author Akash Pallath
# This code is licensed under the MIT license (see LICENSE.txt for details).
"""
Classes to fetch all journal articles in the references of (i.e. backward search)
or citing (i.e. forward search) a set of journal articles.

For backward search, you can use either Crossref or COCI (should be equivalent).
For forward search, you can only use COCI at the moment.
"""
from collections import OrderedDict
import logging
import warnings

from tqdm import tqdm
from stqdm import stqdm

from paperfetcher import GlobalConfig
from paperfetcher.apiclients import CrossrefQuery, COCIQuery
from paperfetcher.content_negotiators import crossref_negotiate_ris
from paperfetcher.datastructures import DOIDataset, RISDataset
from paperfetcher.exceptions import SearchError, ContentNegotiationError, RISParsingError

# Logging
logger = logging.getLogger(__name__)
logger.setLevel(GlobalConfig.loglevel)


class CrossrefBackwardReferenceSearch:
    """
    Retrieves (the DOIs of) all articles in the references of a list of (DOIs of) articles
    by using the Crossref REST API.

    Args:
        search_dois (list): List of DOIs (str) to fetch references of.

    Attributes:
        search_dois (list): List of DOIs (str) to fetch references of.
        result_dois (set): Set of DOIs which are referenced by the DOIs in `search_dois`.

    Example:
        >>> search = snowballsearch.CrossrefBackwardReferenceSearch(["10.1021/acs.jpcb.1c02191", "10.1073/pnas.2018234118"])
        >>> search()
        >>> len(search)
        140
        >>> search.result_dois
        {'10.1021/jp972543+', '10.1073/pnas.0708088105',  ... ,  '10.1073/pnas.0705830104'}
    """
    def __init__(self, search_dois: list):
        self.search_dois = search_dois
        self.result_dois = set()  # prevent duplicates

    # Alternate constructor
    @classmethod
    def from_DOIDataset(cls, search_dataset: DOIDataset):
        """
        Constructs a search object from a DOIDataset.

        Args:
            search_dataset (DOIDataset): Dataset of DOIs to fetch references of.

        Returns:
            CrossrefBackwardReferenceSearch
        """
        return cls(search_dataset._items)

    # Properties
    def __len__(self):
        return (len(self.result_dois))

    @classmethod
    def _check_doi_exists(cls, doi: str):
        """
        Checks if DOI is indexed in Crossref.

        Args:
            doi (str): DOI to check.

        Returns:
            bool
        """
        components = OrderedDict([("works", doi)])
        query = CrossrefQuery(components)
        query()
        return query.response.status_code == 200  # OK?

    @classmethod
    def _check_doi_has_references(cls, doi: str):
        """
        Checks if DOI has references indexed in Crossref.

        Args:
            doi (str): DOI to check.

        Returns:
            bool
        """
        components = OrderedDict([("works", doi)])
        query = CrossrefQuery(components)
        query()

        try:
            data = query.response.json()
            doi_dicts = data['message'].get('reference', None)
            if doi_dicts is None:
                return False
            if len(doi_dicts) == 0:
                return False
            return True

        except Exception:
            warnings.warn("Cannot decode results for the DOI %s" % doi)
            return False  # Cannot decode => has no references

    @classmethod
    def _fetch_all_reference_dois(cls, doi: str):
        """
        Fetches all references of DOI from Crossref.

        Args:
            doi (str): DOI to fetch references of.

        Returns:
            reference_dois (list): List of DOIs

        Raises:
            SearchError if unable to convert query response to JSON format and extract reference data from it.
        """
        components = OrderedDict([("works", doi)])
        query = CrossrefQuery(components)
        query()

        try:
            data = query.response.json()
            doi_dicts = data['message']['reference']

        except Exception:
            raise SearchError("Cannot decode results for the DOI %s" % doi)

        reference_dois = []
        for dict in doi_dicts:
            ref_doi = dict.get("DOI", None)
            if ref_doi is not None:
                reference_dois.append(ref_doi)
            else:
                warnings.warn("In references of %s, reference object %s does not have a DOI field. Skipping this reference." % (doi, dict))  # warn but continue

        return reference_dois

    # Perform search
    def __call__(self):
        if GlobalConfig.streamlit:
            iterable = stqdm(self.search_dois)
        else:
            iterable = tqdm(self.search_dois)

        for doi in iterable:
            # Checks
            if not self._check_doi_exists(doi):
                warnings.warn("DOI %s not indexed in Crossref. Skipping this DOI." % doi)  # warn but continue
                continue  # move on to the next DOI

            if not self._check_doi_has_references(doi):
                warnings.warn("DOI %s does not have reference metadata in Crossref. Skipping this DOI." % doi)  # warn but continue
                continue  # move on to the next DOI

            # Fetch & update results
            try:
                doi_list = self._fetch_all_reference_dois(doi)
                self.result_dois.update(doi_list)

            except SearchError:
                warnings.warn("Error in retrieving reference metadata for DOI %s. Skipping this DOI." % doi)

    def get_DOIDataset(self):
        """
        Returns search results as a DOIDataset object.

        Returns:
            DOIDataset
        """
        return DOIDataset(list(self.result_dois))

    def get_RISDataset(self):
        """
        Returns search results as an RISDataset object. Uses the Crossref REST
        API for content negotation.

        Returns:
            RISDataset
        """
        RIS_dicts = []

        if GlobalConfig.streamlit:
            result_dois = stqdm(self.result_dois, desc="Converting results to RIS format.")
        else:
            result_dois = tqdm(self.result_dois, desc="Converting results to RIS format.")

        for doi in result_dois:
            try:
                # Use Crossref for content negotation
                ris_ref = crossref_negotiate_ris(doi)[0]

            except (ContentNegotiationError, RISParsingError):
                warnings.warn("Failed to get RIS metadata for DOI %s. Appending just the DOI to the RIS dataset." % doi)
                ris_ref = {'type_of_reference': 'JOUR', 'doi': doi}

            # Add to list
            RIS_dicts.append(ris_ref)

        return RISDataset(RIS_dicts)


class COCIBackwardReferenceSearch:
    """
    Retrieves the (DOIs of) all articles in the references of a list of (DOIs of) articles
    by using the COCI REST API.

    Args:
        search_dois (list): List of DOIs (str) to fetch references of.

    Attributes:
        search_dois (list): List of DOIs (str) to fetch references of.
        result_dois (set): Set of DOIs which are referenced by the DOIs in `search_dois`.

    Example:
        >>> search = snowballsearch.COCIBackwardReferenceSearch(["10.1021/acs.jpcb.1c02191", "10.1073/pnas.2018234118"])
        >>> search()
        >>> len(search)
        140
        >>> search.result_dois
        {'10.1021/jp972543+', '10.1073/pnas.0708088105',  ... ,  '10.1073/pnas.0705830104'}
    """
    def __init__(self, search_dois: list):
        self.search_dois = search_dois
        self.result_dois = set()  # prevent duplicates

    # Alternate constructor
    @classmethod
    def from_DOIDataset(cls, search_dataset: DOIDataset):
        """
        Constructs a search object from a DOIDataset.

        Args:
            search_dataset (DOIDataset): Dataset of DOIs to fetch references of.

        Returns:
            COCIBackwardReferenceSearch
        """
        return cls(search_dataset._items)

    # Properties
    def __len__(self):
        return (len(self.result_dois))

    @classmethod
    def _check_doi_exists(cls, doi: str):
        """
        Checks if DOI is indexed in COCI.

        Args:
            doi (str): DOI to check.

        Returns:
            bool
        """
        components = OrderedDict([("references", doi)])
        query = COCIQuery(components)
        query()
        return query.response.status_code == 200

    @classmethod
    def _fetch_all_reference_dois(cls, doi: str):
        """
        Fetches all references of DOI from COCI.

        Args:
            doi (str): DOI to fetch references of.

        Returns:
            reference_dois (list): List of DOIs

        Raises:
            SearchError if unable to convert query response to JSON format and extract reference data from it.
        """
        components = OrderedDict([("references", doi)])
        query = COCIQuery(components)

        query()

        try:
            references = query.response.json()

            reference_dois = []

            for dict in references:
                ref_doi = dict["cited"]
                reference_dois.append(ref_doi)

            return reference_dois

        except Exception:
            raise SearchError("Cannot decode results for the DOI %s" % doi)

    # Perform search
    def __call__(self):
        if GlobalConfig.streamlit:
            iterable = stqdm(self.search_dois)
        else:
            iterable = tqdm(self.search_dois)

        for doi in iterable:
            # Checks
            if not self._check_doi_exists(doi):
                warnings.warn("DOI %s not found in COCI. Skipping this DOI." % doi)  # warn but continue
                continue

            try:
                # Fetch & update results
                doi_list = self._fetch_all_reference_dois(doi)
                self.result_dois.update(doi_list)

            except SearchError:
                warnings.warn("Error in retrieving reference metadata for DOI %s. Skipping this DOI." % doi)

    def get_DOIDataset(self):
        """
        Returns search results as a DOIDataset object.

        Returns:
            DOIDataset
        """
        return DOIDataset(list(self.result_dois))

    def get_RISDataset(self):
        """
        Returns search results as an RISDataset object. Uses the Crossref REST
        API for content negotation.

        Returns:
            RISDataset
        """
        RIS_dicts = []

        if GlobalConfig.streamlit:
            result_dois = stqdm(self.result_dois, desc="Converting results to RIS format.")
        else:
            result_dois = tqdm(self.result_dois, desc="Converting results to RIS format.")

        for doi in result_dois:
            try:
                # Use Crossref for content negotation
                ris_ref = crossref_negotiate_ris(doi)[0]

            except (ContentNegotiationError, RISParsingError):
                warnings.warn("Failed to get RIS metadata for DOI %s. Appending just the DOI to the RIS dataset." % doi)
                ris_ref = {'type_of_reference': 'JOUR', 'doi': doi}

            # Add to list
            RIS_dicts.append(ris_ref)

        return RISDataset(RIS_dicts)


class COCIForwardCitationSearch:
    """
    Retrieves the (DOIs of) all articles citing a list of (DOIs of) articles
    by using the COCI REST API.

    Args:
        search_dois (list): List of DOIs (str) to fetch citations of.

    Attributes:
        search_dois (list): List of DOIs (str) to fetch citations of.
        result_dois (set): Set of DOIs which cite the DOIs in `search_dois`.

    Example:
        >>> search = snowballsearch.COCIForwardCitationSearch(["10.1021/acs.jpcb.8b11423", "10.1073/pnas.2018234118"])
        >>> search()
        >>> len(search)
        11
        >>> search.result_dois
        {'10.1039/c9sc02097g', '10.1021/acs.jpcb.1c05748', ... , '10.1021/acs.jpclett.9b02052'}
    """
    def __init__(self, search_dois: list):
        self.search_dois = search_dois
        self.result_dois = set()  # prevent duplicates

    # Alternate constructor
    @classmethod
    def from_DOIDataset(cls, search_dataset: DOIDataset):
        """
        Constructs a search object from a DOIDataset.

        Args:
            search_dataset (DOIDataset): Dataset of DOIs to fetch references of.

        Returns:
            COCIForwardReferenceSearch
        """
        return cls(search_dataset._items)

    # Properties
    def __len__(self):
        return (len(self.result_dois))

    @classmethod
    def _check_doi_exists(cls, doi: str):
        """
        Checks if DOI is indexed in COCI.

        Args:
            doi (str): DOI to check.

        Returns:
            bool
        """
        components = OrderedDict([("citations", doi)])
        query = COCIQuery(components)
        query()
        return query.response.status_code == 200

    @classmethod
    def _fetch_all_citation_dois(cls, doi: str):
        """
        Fetches all citations of DOI from COCI.

        Args:
            doi (str): DOI to fetch citations of.

        Returns:
            citation_dois (list): List of DOIs

        Raises:
            SearchError if unable to convert query response to JSON format and extract citation data from it.
        """
        components = OrderedDict([("citations", doi)])
        query = COCIQuery(components)
        query()

        try:
            references = query.response.json()

            citation_dois = []

            for dict in references:
                cit_doi = dict["citing"]
                citation_dois.append(cit_doi)

            return citation_dois

        except Exception:
            raise SearchError("Cannot decode results for the DOI %s" % doi)

    # Perform search
    def __call__(self):
        if GlobalConfig.streamlit:
            iterable = stqdm(self.search_dois)
        else:
            iterable = tqdm(self.search_dois)

        for doi in iterable:
            # Checks
            if not self._check_doi_exists(doi):
                warnings.warn("DOI %s not found in COCI. Skipping this DOI." % doi)  # warn but continue
                continue

            try:
                # Fetch & update results
                doi_list = self._fetch_all_citation_dois(doi)
                self.result_dois.update(doi_list)

            except SearchError:
                warnings.warn("Error in retrieving citation metadata for DOI %s. Skipping this DOI." % doi)

    def get_DOIDataset(self):
        """
        Returns search results as a DOIDataset object.

        Returns:
            DOIDataset
        """
        return DOIDataset(list(self.result_dois))

    def get_RISDataset(self):
        """
        Returns search results as an RISDataset object. Uses the Crossref REST
        API for content negotation.

        Returns:
            RISDataset
        """
        RIS_dicts = []

        if GlobalConfig.streamlit:
            result_dois = stqdm(self.result_dois, desc="Converting results to RIS format.")
        else:
            result_dois = tqdm(self.result_dois, desc="Converting results to RIS format.")

        for doi in result_dois:
            try:
                # Use Crossref for content negotation
                ris_ref = crossref_negotiate_ris(doi)[0]

            except (ContentNegotiationError, RISParsingError):
                warnings.warn("Failed to get RIS metadata for DOI %s. Appending just the DOI to the RIS dataset." % doi)
                ris_ref = {'type_of_reference': 'JOUR', 'doi': doi}

            # Add to list
            RIS_dicts.append(ris_ref)

        return RISDataset(RIS_dicts)



# @author Akash Pallath
# This code is licensed under the MIT license (see LICENSE.txt for details).
"""
Integration tests for paperfetcher.snowballsearch package.
"""
import logging
import os

from paperfetcher import snowballsearch

logger = logging.getLogger(__name__)


def test_CrossrefBackward():
    input_DOIs = ["10.1021/acs.jpcb.1c02191", "10.1073/pnas.2018234118"]
    test_output_DOI_members = ["10.1021/acs.jpcb.8b11423"]

    search = snowballsearch.CrossrefBackwardReferenceSearch(input_DOIs)
    search()

    print(len(search))  # should be 140
    assert(len(search) == 140)

    print(search.result_dois)
    for doi in test_output_DOI_members:
        assert(doi in search.result_dois)

    if not os.path.exists("./tmp/"):
        os.makedirs("./tmp/")

    doi_ds = search.get_DOIDataset()

    ris_ds = search.get_RISDataset()

    # Check conversion to text
    doi_ds.save_txt("./tmp/snowball_crossref_back.txt")

    # Check conversion to RIS
    ris_ds.save_ris("./tmp/snowball_crossref_back.ris")


def test_CrossrefBackward_errorhandling():
    input_DOIs = ["10.1021/acs.jpcb.1c02191", "xx.yy.zz/12345.67890", "10.1146/annurev-conmatphys-040220-045516", "10.1073/pnas.2018234118"]
    # xx.yy.zz/12345.67890 does not exist
    # annual reviews does not, at present, index its references in Crossref

    test_output_DOI_members = ["10.1021/acs.jpcb.8b11423"]

    search = snowballsearch.CrossrefBackwardReferenceSearch(input_DOIs)
    search()

    print(len(search))  # should be 140
    assert(len(search) == 140)

    print(search.result_dois)
    for doi in test_output_DOI_members:
        assert(doi in search.result_dois)


def test_COCIBackward():
    input_DOIs = ["10.1021/acs.jpcb.1c02191", "10.1073/pnas.2018234118"]
    test_output_DOI_members = ["10.1021/acs.jpcb.8b11423"]

    search = snowballsearch.COCIBackwardReferenceSearch(input_DOIs)
    search()

    print(len(search))  # should be 140
    assert(len(search) == 140)

    print(search.result_dois)
    for doi in test_output_DOI_members:
        assert(doi in search.result_dois)

    if not os.path.exists("./tmp/"):
        os.makedirs("./tmp/")

    doi_ds = search.get_DOIDataset()

    ris_ds = search.get_RISDataset()

    # Check conversion to text
    doi_ds.save_txt("./tmp/snowball_COCI_back.txt")

    # Check conversion to RIS
    ris_ds.save_ris("./tmp/snowball_COCI_back.ris")


def test_COCIBackward_errorhandling():
    input_DOIs = ["10.1021/acs.jpcb.1c02191", "xx.yy.zz/12345.67890", "10.1146/annurev-conmatphys-040220-045516", "10.1073/pnas.2018234118"]
    # xx.yy.zz/12345.67890 does not exist
    # annual reviews does not, at present, index its references in Crossref or COCI

    test_output_DOI_members = ["10.1021/acs.jpcb.8b11423"]

    search = snowballsearch.COCIBackwardReferenceSearch(input_DOIs)
    search()

    print(len(search))  # should be 140
    assert(len(search) == 140)

    print(search.result_dois)
    for doi in test_output_DOI_members:
        assert(doi in search.result_dois)


def test_COCIForward():
    input_DOIs = ["10.1021/acs.jpcb.8b11423", "10.1073/pnas.2018234118"]
    test_output_DOI_members = ["10.1021/acs.jpcb.1c02191"]

    search = snowballsearch.COCIForwardCitationSearch(input_DOIs)
    search()
    print(len(search))
    print(search.result_dois)
    for doi in test_output_DOI_members:
        assert(doi in search.result_dois)

    if not os.path.exists("./tmp/"):
        os.makedirs("./tmp/")

    doi_ds = search.get_DOIDataset()

    ris_ds = search.get_RISDataset()

    # Check conversion to text
    doi_ds.save_txt("./tmp/snowball_COCI_fwd.txt")

    # Check conversion to RIS
    ris_ds.save_ris("./tmp/snowball_COCI_fwd.ris")


def test_COCIForward_errorhandling():
    input_DOIs = ["10.1021/acs.jpcb.8b11423", "xx.yy.zz/12345.67890", "10.1146/annurev-conmatphys-040220-045516", "10.1073/pnas.2018234118"]
    test_output_DOI_members = ["10.1021/acs.jpcb.1c02191"]

    search = snowballsearch.COCIForwardCitationSearch(input_DOIs)
    search()

    print(len(search))  # this will change with time, do NOT test this value

    print(search.result_dois)
    for doi in test_output_DOI_members:
        assert(doi in search.result_dois)



# @author Akash Pallath
# This code is licensed under the MIT license (see LICENSE.txt for details).
"""Functions to parse data returned from queries.

New parsers can be added here.
"""


def crossref_date_parser(date):
    """Function to parse date.

    Returns:
        str"""
    return "-".join([str(num) for num in date['date-parts'][0]])


def crossref_authors_parser(author_array):
    """Function to parse authors.

    Returns:
        str"""
    try:
        authors = ", ".join(d['family'] for d in author_array)
    except (KeyError, TypeError):
        authors = ""
    return authors


def crossref_title_parser(title):
    """Function to parse title.

    Returns:
        str"""
    return " ".join(title[0].split())



# @author Akash Pallath
# This code is licensed under the MIT license (see LICENSE.txt for details).
"""
Unit test suite for paperfetcher.handsearch package.
"""
import logging

from paperfetcher import handsearch
from paperfetcher.exceptions import SearchError

logger = logging.getLogger(__name__)

############################################################################
# CrossrefSearch unit tests
############################################################################


def test_check_issn_exists():
    result = handsearch.CrossrefSearch._check_issn_exists("1476-4687")
    assert(result)


def test_fetch_count():
    count = handsearch.CrossrefSearch._fetch_count("1476-4687")
    logger.info(count)
    assert(type(count) == int)


def test_fetch_count_errorhandling():
    try:
        handsearch.CrossrefSearch._fetch_count("XXXXX")
    except SearchError as e:
        print(str(e))


def test_fetch_batch():
    data = handsearch.CrossrefSearch._fetch_batch("1476-4687", size=5)
    logger.info(data)
    assert(type(data) == dict)
    assert(len(data['items']) == 5)


def test_fetch_batch_errorhandling():
    try:
        handsearch.CrossrefSearch._fetch_batch("XXXXX", size=5)
    except SearchError as e:
        print(str(e))



# @author Akash Pallath
# This code is licensed under the MIT license (see LICENSE.txt for details).
"""
Custom data structures for paperfetcher.
"""
import contextlib
import csv
import logging

import pandas as pd
import rispy
from rispy.config import LIST_TYPE_TAGS, TAG_KEY_MAPPING

from paperfetcher import GlobalConfig
from paperfetcher.exceptions import DatasetError

# Logging
logger = logging.getLogger(__name__)
logger.setLevel(GlobalConfig.loglevel)


class Dataset:
    """
    Abstract interface that defines functions for child Dataset classes to implement.

    Datasets are designed to store [usually tabular] data (as input to or output from paperfetcher searches), export
    data to pandas DataFrames, and load/save data to disk using common data formats (txt, csv, xlsx).

    Args:
        items (iterable): Items to store in dataset (default=[]).
    """
    def __init__(self, items: list = []):
        self._items = list(items)

    @classmethod
    def from_txt(cls, file):
        """Loads dataset from .txt file."""
        # Child class must implement this.
        raise NotImplementedError()

    @classmethod
    def from_csv(cls, file):
        """Loads dataset from .csv file."""
        # Child class must implement this.
        raise NotImplementedError()

    @classmethod
    def from_excel(cls, file):
        """Loads dataset from Excel file file."""
        # Child class must implement this.
        raise NotImplementedError()

    # Properties
    def __len__(self):
        return len(self._items)

    def __repr__(self):
        return self.__class__.__name__ + " with {} items: ".format(len(self)) + repr(self._items)

    def append(self, item):
        """Adds an item to the dataset."""
        self._items.append(item)

    def extend(self, items: list):
        """Adds each item from a list of items to the dataset."""
        self._items.extend(items)

    # Export to pandas
    def to_df(self):
        """Converts dataset to DataFrame."""
        # Child class must implement this.
        raise NotImplementedError()

    # Export to disk
    def save_txt(self, file):
        """Saves dataset to .txt file."""
        # Child class must implement this.
        raise NotImplementedError()

    def save_csv(self, file):
        """Saves dataset to .csv file."""
        # Child class must implement this.
        raise NotImplementedError()

    def save_excel(self, file):
        """Saves dataset to Excel file."""
        # Child class must implement this.
        raise NotImplementedError()


class DOIDataset(Dataset):
    """
    Stores a dataset of DOIs.

    DOIDatasets can be exported to pandas DataFrames, and loaded from or saved to disk
    in text, CSV, or Excel file formats.

    Args:
        items (list): List of DOIs (str) to store (default=[]).

    Examples:
        To create a DOIDataset object from a list of DOIs:

        >>> ds = DOIDataset(["x1.y1.z1/123123", "x2.y2.z2/456456"])

        To add a DOI to the DOIDataset object:

        >>> ds.append("x3.y3.z3/789789")

        To export the DOIDataset object to a pandas DataFrame:

        >>> df = ds.to_df()
        >>> df
                       DOI
        0  x1.y1.z1/123123
        1  x2.y2.z2/456456
        3  x3.y3.z3/789789

        To save data to disk:

        >>> ds.save_txt("dois.txt")
        >>> ds.save_csv("dois.csv")
        >>> ds.save_excel("dois.xlsx")
    """
    def __init__(self, items: list = []):
        super().__init__(items)

    def extend_dataset(self, ds: 'DOIDataset'):
        """Appends all items from DOIDataset ds to the end of the current dataset."""
        self.extend(ds._items)

    def to_df(self):
        """Converts dataset to DataFrame."""
        return pd.DataFrame(self._items, columns=['DOI'])

    def to_txt_string(self):
        """Returns a string which can be written to .txt file"""
        txt = ""
        for doi in self._items:
            txt = txt + doi + "\n"
        return txt

    def save_txt(self, file):
        """Saves dataset to .txt file."""
        if hasattr(file, 'write'):
            file_ctx = contextlib.nullcontext(file)
        else:
            if not file.endswith('.txt'):
                file = file + '.txt'
            file_ctx = open(file, "w")

        with file_ctx as f:
            f.write(self.to_txt_string())

    def save_csv(self, file):
        """Saves dataset to .csv file."""
        if hasattr(file, 'write'):
            file_ctx = contextlib.nullcontext(file)
        else:
            if not file.endswith('.csv'):
                file = file + '.csv'
            file_ctx = open(file, "w")

        with file_ctx as f:
            write = csv.writer(f)
            write.writerow(["DOI"])
            write.writerows([[item] for item in self._items])

    def save_excel(self, file):
        """Saves dataset to Excel file."""
        if not file.endswith('.xlsx'):
            file = file + '.xlsx'
        df = pd.DataFrame(self._items, columns=['DOI'])
        df.to_excel(file)


class CitationsDataset(Dataset):
    """
    Stores a tabular dataset of citations, with multiple custom fields.

    CitationsDatasets can be exported to pandas DataFrames, and loaded from or saved to disk
    in text, CSV, or Excel file formats.

    Args:
        field_names (tuple): Names (str) of fields.
        items (list): List of citations to store (default=[]). Each citation should be an iterable of length len(field_names).

    Examples:
        To create a CitationsDataset object to store the DOI, URL, article title, authors, and date of issue for each citation:

        >>> field_names = ["DOI", "URL", "title", "author", "issued"]
        >>> data = [["10.xxyy/0.0.0.000001", "https://dx.doi.org/10.xxyy/0.0.0.000001", "A study of A", "P, Q and R", "2020-02-20"],
        ...         ["10.xxyy/0.0.0.000002", "https://dx.doi.org/10.xxyy/0.0.0.000002", "An investigation into B", "Q, R and P", "2020-03-20"],
        ...         ["10.xxyy/0.0.0.000003", "https://dx.doi.org/10.xxyy/0.0.0.000003", "The causes of C in D", "R, P and Q", "2020-04-20"],
        ...         ["10.xxyy/0.0.0.000004", "https://dx.doi.org/10.xxyy/0.0.0.000004", "Characterizing the role of E on F", "P and Q", "2020-05-20"]]
        >>> ds = CitationsDataset(field_names, data)

        To add a citation to the CitationsDataset:

        >>> ds.append(["10.xxyy/0.0.0.000005", "https://dx.doi.org/10.xxyy/0.0.0.000005", "The G effect", "Q and P", "2020-06-20"])

        To export the DOIDataset object to a pandas DataFrame:

        >>> df = ds.to_df()
        >>> df
                            DOI                                      URL                              title      author      issued
        0  10.xxyy/0.0.0.000001  https://dx.doi.org/10.xxyy/0.0.0.000001                       A study of A  P, Q and R  2020-02-20
        1  10.xxyy/0.0.0.000002  https://dx.doi.org/10.xxyy/0.0.0.000002            An investigation into B  Q, R and P  2020-03-20
        2  10.xxyy/0.0.0.000003  https://dx.doi.org/10.xxyy/0.0.0.000003               The causes of C in D  R, P and Q  2020-04-20
        3  10.xxyy/0.0.0.000004  https://dx.doi.org/10.xxyy/0.0.0.000004  Characterizing the role of E on F     P and Q  2020-05-20
        4  10.xxyy/0.0.0.000005  https://dx.doi.org/10.xxyy/0.0.0.000005                       The G effect     Q and P  2020-06-20

        To save data to disk:

        >>> ds.save_txt("cits.txt")
        >>> ds.save_csv("cits.csv")
        >>> ds.save_excel("cits.xlsx")
    """
    def __init__(self, field_names: tuple, items: list = []):
        super().__init__(items)
        self.field_names = field_names
        self.num_fields = len(field_names)
        for itemidx, item in enumerate(self._items):
            if len(item) != self.num_fields:
                raise DatasetError("Item at index %d is of incorrect length." % itemidx)

    def append(self, item):
        """Adds a citation to the dataset."""
        if len(item) != self.num_fields:
            raise DatasetError("Item is of incorrect length.")
        super().append(item)

    def extend(self, items):
        """Adds each citation from a list of citations (i.e. eacher inner list of nested list) to the dataset."""
        for itemidx, item in enumerate(items):
            if len(item) != self.num_fields:
                raise DatasetError("Item at index %d is of incorrect length." % itemidx)
        super().extend(items)

    def to_df(self):
        """Converts dataset to DataFrame."""
        return pd.DataFrame(self._items, columns=self.field_names)

    def save_txt(self, file):
        """Saves dataset to .txt file."""
        if hasattr(file, 'write'):
            file_ctx = contextlib.nullcontext(file)
        else:
            if not file.endswith('.txt'):
                file = file + '.txt'
            file_ctx = open(file, "w")

        with file_ctx as f:
            f.write("\t".join(self.field_names) + "\n")
            for item in self._items:
                f.write("\t".join(item) + "\n")

    def save_csv(self, file):
        """Saves dataset to .csv file. The first row of the CSV file contains field names."""
        if hasattr(file, 'write'):
            file_ctx = contextlib.nullcontext(file)
        else:
            if not file.endswith('.csv'):
                file = file + '.csv'
            file_ctx = open(file, "w")

        with file_ctx as f:
            write = csv.writer(f)
            write.writerow(self.field_names)
            write.writerows(self._items)

    def save_excel(self, file):
        """Saves dataset to Excel file (uses Pandas). The first row of the Excel file contains field names."""
        if not file.endswith('.xlsx'):
            file = file + '.xlsx'
        df = pd.DataFrame(self._items, columns=self.field_names)
        df.to_excel(file)


# rispy modification to remove header containing reference number from RIS output
class HeadlessRISWriter(rispy.writer.BaseWriter):
    START_TAG = "TY"
    PATTERN = "{tag}  - {value}"
    DEFAULT_MAPPING = TAG_KEY_MAPPING
    DEFAULT_LIST_TAGS = LIST_TYPE_TAGS
    DEFAULT_DELIMITER_MAPPING = {"UR": ";"}
    

class RISDataset(Dataset):
    """
    Stores a dataset of RIS items. RIS items are rispy-readable dictionaries.

    An RISDataset can be created from an RIS-formatted string, or from an RIS file.
    An RISDataset can be written to an RIS-formatted string, or to an RIS file.
    Individual item dictionaries in an RISDataset can be modified to add new tags or change
    the values of existing tags.

    Args:
        items (list): List of citations to store (default=[]). Each citation should be a rispy-readable dictionary.

    Examples:
        To create an RISDataset from a list of rispy-readable dictionaries (see rispy doc on GitHub for details):

        >>> dict_list = [{'journal_name': ...}, {'journal_name: ...'}, ...]
        >>> ds = RISDataset(dict_list)

        To load an RISDataset from an RIS-formatted string:
        >>> ds = RISDataset.from_ris_string(ris_string)

        To load an RISDataset from an RIS file:
        >>> ds = RISDataset.from_ris(ris_file)
    """
    def __init__(self, items: list = []):
        super().__init__(items)

    @classmethod
    def from_ris_string(cls, ris_string):
        """Loads dataset from RIS-formatted string."""
        items = rispy.loads(ris_string)
        return cls(items)

    @classmethod
    def from_ris(cls, file):
        """Loads dataset from RIS file."""
        with open(file, 'r') as f:
            items = rispy.load(f)
        return cls(items)

    def extend_dataset(self, ds: 'RISDataset'):
        """Appends all items from RISDataset ds to the end of the current dataset."""
        self.extend(ds._items)

    def to_ris_string(self, headers=False):
        """Returns a string which can be written to .ris file.

        Args:
            headers (bool, default=False): If set to true, writes reference number before each RIS entry."""
        if headers:
            return rispy.dumps(self._items)
        else:
            return rispy.dumps(self._items, implementation=HeadlessRISWriter)

    def save_ris(self, filename, headers=False):
        """Saves dataset to .ris file.

        Args:
            filename: Path to file to write RIS data to.
            headers (bool, default=False): If set to true, writes reference number before each RIS entry."""
        if headers:
            with open(filename, 'w') as f:
                rispy.dump(self._items, f)
        else:
            with open(filename, 'w') as f:
                rispy.dump(self._items, f, implementation=HeadlessRISWriter)



# @author Akash Pallath
# This code is licensed under the MIT license (see LICENSE.txt for details).
"""
Unit test suite for paperfetcher.datastructures package.
"""
import os

from pathlib import Path
import pytest

from paperfetcher.datastructures import DOIDataset, CitationsDataset, RISDataset


"""DOIDataset tests"""


@pytest.fixture
def doids():
    data = ["10.xxyy/0.0.0.000001",
            "10.xxyy/0.0.0.000002",
            "10.xxyy/0.0.0.000003",
            "10.xxyy/0.0.0.000004",
            "10.xxyy/0.0.0.000005"]
    return DOIDataset(data)


@pytest.fixture
def citds():
    field_names = ["DOI", "URL", "title", "author", "issued"]
    data = [["10.xxyy/0.0.0.000001", "https://dx.doi.org/10.xxyy/0.0.0.000001", "A study of A", "P, Q and R", "2020-02-20"],
            ["10.xxyy/0.0.0.000002", "https://dx.doi.org/10.xxyy/0.0.0.000002", "An investigation into B", "Q, R and P", "2020-03-20"],
            ["10.xxyy/0.0.0.000003", "https://dx.doi.org/10.xxyy/0.0.0.000003", "The causes of C in D", "R, P and Q", "2020-04-20"],
            ["10.xxyy/0.0.0.000004", "https://dx.doi.org/10.xxyy/0.0.0.000004", "Characterizing the role of E on F", "P and Q", "2020-05-20"],
            ["10.xxyy/0.0.0.000005", "https://dx.doi.org/10.xxyy/0.0.0.000005", "The G effect", "Q and P", "2020-06-20"]]
    return CitationsDataset(field_names, data)


@pytest.fixture
def ris_entry():
    entry = "TY  - JOUR\nDO  - 10.1073/pnas.2018234118\nUR  - http://dx.doi.org/10.1073/pnas.2018234118\nTI  - Identifying hydrophobic protein patches to inform protein interaction interfaces\nT2  - Proceedings of the National Academy of Sciences\nAU  - Rego, Nicholas B.\nAU  - Xi, Erte\nAU  - Patel, Amish J.\nPY  - 2021\nDA  - 2021/02/01\nPB  - Proceedings of the National Academy of Sciences\nSP  - e2018234118\nIS  - 6\nVL  - 118\nSN  - 1091-6490\nER  -"
    return entry


@pytest.fixture
def ris_file():
    path = Path(__file__).parent / "input_data/ris1.ris"
    return path


def test_DOIDataset_to_df(doids):
    df = doids.to_df()
    print(df)


def test_DOIDataset_save_txt(doids):
    if not os.path.exists("./tmp/"):
        os.makedirs("./tmp/")
    doids.save_txt("./tmp/doiunit.txt")


def test_DOIDataset_save_csv(doids):
    if not os.path.exists("./tmp/"):
        os.makedirs("./tmp/")
    doids.save_csv("./tmp/doiunit.csv")


def test_DOIDataset_save_excel(doids):
    if not os.path.exists("./tmp/"):
        os.makedirs("./tmp/")
    doids.save_excel("./tmp/doiunit.xlsx")


"""CitationsDataset tests"""


def test_CitationsDataset_to_df(citds):
    df = citds.to_df()
    print(df)


def test_CitationsDataset_save_txt(citds):
    if not os.path.exists("./tmp/"):
        os.makedirs("./tmp/")
    citds.save_txt("./tmp/citunit.txt")


def test_CitationsDataset_save_csv(citds):
    if not os.path.exists("./tmp/"):
        os.makedirs("./tmp/")
    citds.save_csv("./tmp/citunit.csv")


def test_CitationsDataset_save_xlsx(citds):
    if not os.path.exists("./tmp/"):
        os.makedirs("./tmp/")
    citds.save_excel("./tmp/citunit.xlsx")


"""RISDataset tests"""


def test_RISDataset_constructors(ris_entry, ris_file):
    # Test constructor to load dataset from RIS-formatted string
    ds1 = RISDataset.from_ris_string(ris_entry)
    print(ds1)
    assert(len(ds1) == 1)

    # Test constructor to load dataset from RIS file
    ds2 = RISDataset.from_ris(ris_file)
    print(ds2)
    assert(len(ds2) == 2)


def test_RISDataset_ris_output(ris_entry):
    # Test RIS reconstruction
    ds = RISDataset.from_ris_string(ris_entry)
    ris_str = RISDataset.to_ris_string(ds)
    print(ris_entry)
    print("---")
    print(ris_str)
    assert(ris_entry.strip() == ris_str.strip())


def test_RISDataset_save_ris(ris_entry):
    if not os.path.exists("./tmp/"):
        os.makedirs("./tmp/")
    ds = RISDataset.from_ris_string(ris_entry)
    ds.save_ris("./tmp/risunit.ris")



# @author Akash Pallath
# This code is licensed under the MIT license (see LICENSE.txt for details).
"""
Unit test suite for paperfetcher.handsearch package.
"""
import logging
import sys

from paperfetcher import handsearch
from paperfetcher.exceptions import ContentNegotiationError

logger = logging.getLogger(__name__)

############################################################################
# Unit tests
############################################################################


def test_crossref_ris():
    data = handsearch.crossref_negotiate_ris(doi="10.1073/pnas.2018234118")
    print(data)


def test_crossref_ris_errorhandling():
    try:
        handsearch.crossref_negotiate_ris(doi="xx.yy.xx/1020304050")
    except ContentNegotiationError:
        return True



# @author Akash Pallath
# This code is licensed under the MIT license (see LICENSE.txt for details).
"""
Unit test suite for paperfetcher.apiclients package.
"""
import logging
import sys

from paperfetcher.apiclients import Query, CrossrefQuery, QueryError

logger = logging.getLogger(__name__)


"""Query tests"""


def test_query():
    query = Query("https://api.github.com")
    query()
    logger.info(query.response.text)


def test_query_fail():
    query = Query("https://www.google.google")
    try:
        query()
    except QueryError:
        return True
    else:
        return False


"""CrossrefQuery tests"""


def test_crossref_query():
    query = CrossrefQuery()
    query.query_base += "works"
    query()
    logger.info(query.response.text)


def test_json():
    query = CrossrefQuery()
    query.query_base += "works"
    query()
    logger.info(query.response.json())



# @author Akash Pallath
# This code is licensed under the MIT license (see LICENSE.txt for details).
"""
Integration tests for paperfetcher.handsearch package.
"""
import logging
import os

from paperfetcher import handsearch
from paperfetcher import parsers

logger = logging.getLogger(__name__)


def test_Crossref_JACS_nokeywords():
    search = handsearch.CrossrefSearch(ISSN="1520-5126", from_date="2020-01-01",
                                       until_date="2020-04-01")
    # fast/low memory search
    search(select=True, select_fields=['DOI'])

    # Check raw
    for idx, result in enumerate(search.results):
        logger.info("{}/{}".format(idx + 1, len(search)))
        logger.info(result)
    print(len(search))
    assert(len(search) == 772)


def test_Crossref_JACS_emptykeywords():
    search = handsearch.CrossrefSearch(ISSN="1520-5126", keyword_list=[""], from_date="2020-01-01",
                                       until_date="2020-04-01")
    # fast/low memory search
    search(select=True, select_fields=['DOI'])

    # Check raw
    for idx, result in enumerate(search.results):
        logger.info("{}/{}".format(idx + 1, len(search)))
        logger.info(result)
    print(len(search))
    assert(len(search) == 772)


def test_Crossref_JACS_hydration_DOIDataset():
    search = handsearch.CrossrefSearch(ISSN="1520-5126", keyword_list=["hydration"], from_date="2018-01-01",
                                       until_date="2020-01-01")
    search()

    # Check raw
    for idx, result in enumerate(search.results):
        logger.info("{}/{}".format(idx + 1, len(search)))
        logger.info(result)
    assert(len(search) == 13)

    # Check DOIDataset
    ds = search.get_DOIDataset()
    assert(len(ds) == 13)
    assert("10.1021/jacs.8b11448" in ds._items)

    # Check conversion to DataFrame
    df = ds.to_df()
    print(df)

    if not os.path.exists("./tmp/"):
        os.makedirs("./tmp/")

    # Check conversion to text
    df = ds.save_txt("./tmp/JACS_hydration_DOIs.txt")
    print(df)

    # Check conversion to csv
    df = ds.save_csv("./tmp/JACS_hydration_DOIs.csv")
    print(df)

    # Check conversion to xlsx
    df = ds.save_excel("./tmp/JACS_hydration_DOIs.xlsx")
    print(df)


def test_Crossref_JACS_hydration_CitationsDataset():
    search = handsearch.CrossrefSearch(ISSN="1520-5126", keyword_list=["hydration"], from_date="2018-01-01",
                                       until_date="2020-01-01")
    search(select=True, select_fields=['DOI', 'URL', 'title', 'author', 'issued'])

    # Check raw
    for idx, result in enumerate(search.results):
        logger.info("{}/{}".format(idx + 1, len(search)))
        logger.info(result)
    assert(len(search) == 13)

    # Check DOIDataset
    ds = search.get_DOIDataset()
    assert(len(ds) == 13)
    assert("10.1021/jacs.8b11448" in ds._items)

    # Check CitationsDataset
    ds = search.get_CitationsDataset(field_list=['DOI', 'URL', 'title', 'author', 'issued'],
                                     field_parsers_list=[None, None, parsers.crossref_title_parser,
                                                         parsers.crossref_authors_parser, parsers.crossref_date_parser])
    print(ds._items)
    assert(len(ds) == 13)

    # Check conversion to DataFrame
    df = ds.to_df()
    print(df)

    if not os.path.exists("./tmp/"):
        os.makedirs("./tmp/")

    # Check conversion to text
    ds.save_txt("./tmp/JACS_hydration.txt")

    # Check conversion to csv
    ds.save_csv("./tmp/JACS_hydration.csv")

    # Check conversion to xlsx
    ds.save_excel("./tmp/JACS_hydration.xlsx")


def test_Crossref_PNAS_hydrophobic_RISDataset():
    search = handsearch.CrossrefSearch(ISSN="1091-6490", keyword_list=["hydrophobic"], from_date="2020-01-01",
                                       until_date="2021-05-01")
    search(select=True, select_fields=["DOI", "abstract"])

    # Check raw
    for idx, result in enumerate(search.results):
        logger.info("{}/{}".format(idx + 1, len(search)))
        logger.info(result)
    assert(len(search) == 7)

    # RISDataset without abstract
    ds_noabs = search.get_RISDataset()
    print(ds_noabs.to_ris_string())

    # RISDataset with abstract
    ds_abs = search.get_RISDataset(extra_field_list=["abstract"],
                                   extra_field_parser_list=[None],
                                   extra_field_rispy_tags=["notes_abstract"])
    print(ds_abs.to_ris_string())

    if not os.path.exists("./tmp/"):
        os.makedirs("./tmp/")

    # Check conversion to RIS without abstracts
    ds_noabs.save_ris("./tmp/PNAS_hydrophobic.ris")

    # Check conversion to RIS with abstracts
    ds_abs.save_ris("./tmp/PNAS_hydrophobic_withabstract.ris")


def test_Crossref_JACS_nomatch_keywords():
    search = handsearch.CrossrefSearch(ISSN="1520-5126", keyword_list=["adsjkfhadjklhfjkahserkjhajsdhrf"], from_date="2020-01-01",
                                       until_date="2020-04-01")

    # fast/low memory search
    search(select=True, select_fields=['DOI', 'URL', 'title', 'author', 'issued', 'abstract'])

    print(len(search))
    assert(len(search) == 0)

    # Check DOIDataset
    ds = search.get_DOIDataset()
    print(len(ds))
    assert(len(ds) == 0)

    # Check citations dataset
    ds = search.get_CitationsDataset(field_list=['DOI', 'URL', 'title', 'author', 'issued'],
                                     field_parsers_list=[None, None, parsers.crossref_title_parser,
                                                         parsers.crossref_authors_parser, parsers.crossref_date_parser])
    print(len(ds))
    assert(len(ds) == 0)

    # RISDataset without abstract
    ds_noabs = search.get_RISDataset()
    print(ds_noabs.to_ris_string())
    print(len(ds_noabs))
    assert(len(ds_noabs) == 0)

    # RISDataset with abstract
    ds_abs = search.get_RISDataset(extra_field_list=["abstract"],
                                   extra_field_parser_list=[None],
                                   extra_field_rispy_tags=["notes_abstract"])
    print(ds_abs.to_ris_string())
    print(len(ds_abs))
    assert(len(ds_abs) == 0)



# @author Akash Pallath
# This code is licensed under the MIT license (see LICENSE.txt for details).
"""
Definitions of all Exceptions raised by paperfetcher.
"""


class QueryError(Exception):
    """Exception raised when query fails."""
    pass


class SearchError(Exception):
    """Exception raised when search fails."""
    pass


class ContentNegotiationError(Exception):
    """Exception raised when content negotiation fails."""
    pass


class RISParsingError(Exception):
    """Exception raised when RIS parsing fails."""
    pass


class DatasetError(Exception):
    """Exception raised when an incorrect operation is performed on a dataset."""
    pass



# @author Akash Pallath
# This code is licensed under the MIT license (see LICENSE.txt for details).
"""
Unit test suite for paperfetcher.snowballsearch package.
"""
import logging

from paperfetcher import snowballsearch
from paperfetcher.datastructures import DOIDataset
from paperfetcher.exceptions import SearchError

logger = logging.getLogger(__name__)

############################################################################
# CrossrefBackwardReferenceSearch unit tests
############################################################################


def test_crback_check_doi_exists():
    test_1 = snowballsearch.CrossrefBackwardReferenceSearch._check_doi_exists("10.1021/acs.jpcb.1c02191")
    test_2 = snowballsearch.CrossrefBackwardReferenceSearch._check_doi_exists("xx.yy.zz/pqr123")
    assert(test_1)
    assert(not test_2)


def test_crback_check_doi_has_references():
    test_1 = snowballsearch.CrossrefBackwardReferenceSearch._check_doi_has_references("10.1021/acs.jpcb.1c02191")
    test_2 = snowballsearch.CrossrefBackwardReferenceSearch._check_doi_has_references("xx.yy.zz/pqr123")
    assert(test_1)
    assert(not test_2)


def test_crback_from_dataset():
    test_ds = DOIDataset(["10.1021/acs.jpcb.1c02191"])
    test = snowballsearch.CrossrefBackwardReferenceSearch.from_DOIDataset(test_ds)
    assert(len(test.search_dois) == 1)
    assert("10.1021/acs.jpcb.1c02191" in test.search_dois)


def test_crback_fetch_all_reference_dois_errorhandling():
    try:
        snowballsearch.CrossrefBackwardReferenceSearch._fetch_all_reference_dois("xx.yy.zz/pqr123")
    except SearchError:
        return True


def test_crback_get_RISDataset():
    search = snowballsearch.CrossrefBackwardReferenceSearch([""])
    search.result_dois = ["10.1021/acs.jpcb.1c02191"]
    print(search.get_RISDataset())


def test_crback_get_RISDataset_errorhandling():
    search = snowballsearch.CrossrefBackwardReferenceSearch([""])
    search.result_dois = ["xx.yy.zz/pqr123"]
    print(search.get_RISDataset())


############################################################################
# COCIBackwardReferenceSearch unit tests
############################################################################


def test_cociback_check_doi_exists():
    test_1 = snowballsearch.COCIBackwardReferenceSearch._check_doi_exists("10.1021/acs.jpcb.1c02191")
    test_2 = snowballsearch.COCIBackwardReferenceSearch._check_doi_exists("xx.yy.zz/pqr123")
    assert(test_1)
    assert(not test_2)


def test_cociback_from_dataset():
    test_ds = DOIDataset(["10.1021/acs.jpcb.1c02191"])
    test = snowballsearch.COCIBackwardReferenceSearch.from_DOIDataset(test_ds)
    assert(len(test.search_dois) == 1)
    assert("10.1021/acs.jpcb.1c02191" in test.search_dois)


def test_cociback_fetch_all_reference_dois_errorhandling():
    try:
        snowballsearch.COCIBackwardReferenceSearch._fetch_all_reference_dois("xx.yy.zz/pqr123")
    except SearchError:
        return True


def test_cociback_get_RISDataset():
    search = snowballsearch.COCIBackwardReferenceSearch([""])
    search.result_dois = ["10.1021/acs.jpcb.1c02191"]
    print(search.get_RISDataset())


def test_cociback_get_RISDataset_errorhandling():
    search = snowballsearch.COCIBackwardReferenceSearch([""])
    search.result_dois = ["xx.yy.zz/pqr123"]
    print(search.get_RISDataset())


############################################################################
# COCIForwardCitationSearch unit tests
############################################################################


def test_cocifwd_check_doi_exists():
    test_1 = snowballsearch.COCIForwardCitationSearch._check_doi_exists("10.1021/acs.jpcb.1c02191")
    test_2 = snowballsearch.COCIForwardCitationSearch._check_doi_exists("xx.yy.zz/pqr123")
    assert(test_1)
    assert(not test_2)


def test_cocifwd_from_dataset():
    test_ds = DOIDataset(["10.1021/acs.jpcb.1c02191"])
    test = snowballsearch.COCIForwardCitationSearch.from_DOIDataset(test_ds)
    assert(len(test.search_dois) == 1)
    assert("10.1021/acs.jpcb.1c02191" in test.search_dois)


def test_cocifwd_fetch_all_citation_dois_errorhandling():
    try:
        snowballsearch.COCIForwardCitationSearch._fetch_all_citation_dois("xx.yy.zz/pqr123")
    except SearchError:
        return True


def test_cocifwd_get_RISDataset():
    search = snowballsearch.COCIForwardCitationSearch([""])
    search.result_dois = ["10.1021/acs.jpcb.1c02191"]
    print(search.get_RISDataset())


def test_cocifwd_get_RISDataset_errorhandling():
    search = snowballsearch.COCIForwardCitationSearch([""])
    search.result_dois = ["xx.yy.zz/pqr123"]
    print(search.get_RISDataset())



# @author Akash Pallath
# This code is licensed under the MIT license (see LICENSE.txt for details).
"""
Client implementations to communicate with various APIs.
"""
from collections import OrderedDict
import logging
import requests
import sys

from paperfetcher import GlobalConfig
from paperfetcher.exceptions import QueryError

# Logging
logger = logging.getLogger(__name__)
logger.setLevel(GlobalConfig.loglevel)


class Query:
    """
    Base class for structuring and executing HTTP GET queries.

    Args:
        base_url (str): Base URL for query (such as api.xyz.com/get).
        query_params (dict): Dictionary of query parameters.
        headers (dict): Dictionary of HTTP headers to pass along with the query.

    Attributes:
        query_base (str): Base URL for query (such as api.xyz.com/get).
        query_params (dict): Dictionary of query parameters.
        headers (dict): Dictionary of HTTP headers.
        response (requests.Response): Response recieved on executing GET query.

    Examples:
        A simple Query to the Github REST API:

        >>> query = Query("https://api.github.com")
        >>> query()
        >>> query.response
        <Response [200]>

        A Query to the Github REST API to fetch a list of all public repositories in the paperfetcher organization:

        >>> query = Query("https://api.github.com/orgs/paperfetcher/repos",
        ...               query_params={"type": "public"},
        ...               headers={"Accept": "application/vnd.github.v3+json"})
        >>> query()
        >>> query.response
        <Response [200]>
    """
    def __init__(self, base_url=None, query_params: dict = {}, headers: str = {}):
        self.query_base = base_url
        self.query_params = query_params
        self.headers = headers
        # Output
        self.__response = None

    @property
    def response(self):
        if self.__response is not None:
            return self.__response
        else:
            raise RuntimeError("Need to run Query first before accessing its response.")

    def _log_request(self, response, *args, **kwargs):
        logger.debug("\n-----Request-----\nMethod: {}\nURL: {}\nBody: {}\nHeaders: {}\n-----------------\n".
                     format(response.request.method,
                            response.request.url,
                            response.request.body,
                            response.request.headers))

    def __call__(self):
        try:
            self.__response = requests.get(self.query_base, params=self.query_params,
                                           headers=self.headers, hooks={'response': self._log_request})

        except requests.exceptions.ConnectionError as e:
            raise QueryError("Unable to run query, could not reach server:" + str(e)).with_traceback(sys.exc_info()[2])

        except requests.exceptions.Timeout as e:
            raise QueryError("Unable to run query, request timed out:" + str(e)).with_traceback(sys.exc_info()[2])

        except requests.exceptions.HTTPError as e:
            raise QueryError("Unable to run query, HTTP error:" + str(e)).with_traceback(sys.exc_info()[2])

        except requests.exceptions.RequestException as e:
            raise QueryError("Unable to run query, exception:" + str(e)).with_traceback(sys.exc_info()[2])


class CrossrefQuery(Query):
    """
    Class for structuring and executing Crossref REST API queries.

    Query components can be added to the base URL by passing an ordered dictionary to the components argument.
    For example, the components dictionary `{"comp1-key": "comp1-value", "comp2": None}` changes
    the query URL to `https://api.crossref.org/comp1-key/comp1-value/comp2/`.

    Args:
        components (collections.OrderedDict): Components to append to the base URL.
        query_params (collections.OrderedDict): Ordered dictionary of query parameters.

    Attributes:
        components (collections.OrderedDict): Components to append to the base URL.
        query_base (str): Base URL for query (https://api.crossref.org/...).
        query_params (collections.OrderedDict): Dictionary of query parameters.
        headers (dict): Dictionary of HTTP headers.
        response (requests.Response): Response recieved on executing GET query to the Crossref API.

    Examples:
        Querying the metadata of a paper with a known DOI:

        >>> query = CrossrefQuery(components={"works": "10.1021/acs.jpcb.1c02191"})
        >>> query()
        >>> query.response
        <Response [200]>
        >>> query.response.json()
        {'status': 'ok', 'message-type': 'work', 'message-version': '1.0.0',
        'message': {...}}

        Query to fetch all articles from a journal with a known ISSN:

        >>> components = OrderedDict([("journals", "1520-5126"),
                                  ("works", None)])
        >>> query = CrossrefQuery(components)
        >>> query()
        >>> query.response
        <Response [200]>
        >>> query.response.json()
        {'status': 'ok', 'message-type': 'work', 'message-version': '1.0.0',
        'message': {...}}
    """

    # Which version of the Crossref API to use.
    __API_VERSION = 1

    def __init__(self, components=OrderedDict(), query_params=OrderedDict()):
        # Crossref API etiquette
        headers = {'User-Agent': GlobalConfig.crossref_useragent}

        # Option to use Crossref Plus
        if GlobalConfig.crossref_plus:
            headers["Crossref-Plus-API-Token"] = GlobalConfig.crossref_plus_auth_token

        super().__init__("https://api.crossref.org/v{}/".format(self.__API_VERSION),
                         query_params=query_params,
                         headers=headers)

        # Specific to Crossref API call structure
        self.components = components

        # The order in which components and query params are added is preserved (this is important!).
        # Both components and query params will be unpacked only at call time.

    def __call__(self):
        # Unpack components
        for k, v in self.components.items():
            if v is not None:
                self.query_base += "%s/%s/" % (k, v)
            else:
                self.query_base += "%s/" % k

        # Call
        super().__call__()


class COCIQuery(Query):
    """
    Class for structuring and executing COCI REST API queries.

    Args:
        components (collections.OrderedDict): Components to append to the base URL.
        query_params (collections.OrderedDict): Ordered dictionary of query parameters.

    Attributes:
        components (collections.OrderedDict): Components to append to the base URL.
        query_base (str): Base URL for query (https://opencitations.net/index/coci/api/v{}/...).
        query_params (collections.OrderedDict): Dictionary of query parameters.
        headers (dict): Dictionary of HTTP headers.
        response (requests.Response): Response recieved on executing GET query to the COCI API.

    Examples:
        Querying the references of a paper with a known DOI:

        >>> query = COCIQuery(components=OrderedDict([("references", "10.1021/acs.jpcb.1c02191")]))
        >>> query()
        >>> query.response
        <Response [200]>
        >>> query.response.json()
        [{'oci': '020010002013610122837192512113701120002010901-0200100000236191212370201090809', 'creation': '2021-05-12', 'timespan': 'P9Y5M19D',
        ...}]

        Querying the citations of a paper with a known DOI:

        >>> query = COCIQuery(components=OrderedDict([("citations", "10.1021/acs.jpcb.1c02191")]))
        >>> query()
        >>> query.response
        <Response [200]>
        >>> query.response.json()
        [{'oci': '0200100000236252421370200020100050206-020010002013610122837192512113701120002010901', 'creation': '2021-10-02', 'timespan': 'P4M20D',
         'journal_sc': 'no', 'author_sc': 'no', 'citing': '10.1002/pol.20210526', 'cited': '10.1021/acs.jpcb.1c02191'}]
    """

    # Which version of the COCI  API to use.
    __API_VERSION = 1

    def __init__(self, components=OrderedDict(), query_params=OrderedDict()):
        super().__init__("https://opencitations.net/index/coci/api/v{}/".format(self.__API_VERSION),
                         query_params=query_params,
                         headers={})

        # Specific to COCI API call structure
        self.components = components

        # The order in which components and query params are added is preserved (this is important!).
        # Both components and query params will be unpacked only at call time.

    def __call__(self):
        # Unpack components
        for k, v in self.components.items():
            if v is not None:
                self.query_base += "%s/%s/" % (k, v)
            else:
                self.query_base += "%s/" % k

        # COCI-specific error handling
        if self.query_base[-1] == "/":
            self.query_base = self.query_base[:-1]

        # Call
        super().__call__()



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("VERSION", "r") as vh:
    version = vh.read()
version = version.strip()

setup(name='paperfetcher',
      python_requires='>3.7',
      version=version,
      author='Akash Pallath',
      author_email='apallath@seas.upenn.edu',
      description='Python package to mine papers for systematic reviews.',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/paperfetcher/paperfetcher',
      packages=['paperfetcher'],
      install_requires=[
          'openpyxl',
          'pandas',
          'pytest',
          'requests',
          'rispy',
          'stqdm',
          'tqdm',
      ],)



# @author Akash Pallath
# This code is licensed under the MIT license (see LICENSE.txt for details).
"""
Classes to fetch all journal works (articles) matching a set of keywords and
within a given date range by querying various APIs.
"""

from collections import OrderedDict
import logging
import warnings

from tqdm import tqdm
from stqdm import stqdm

from paperfetcher import GlobalConfig
from paperfetcher.apiclients import CrossrefQuery
from paperfetcher.content_negotiators import crossref_negotiate_ris
from paperfetcher.datastructures import DOIDataset, CitationsDataset, RISDataset
from paperfetcher.exceptions import SearchError

# Logging
logger = logging.getLogger(__name__)
logger.setLevel(GlobalConfig.loglevel)


class CrossrefSearch:
    """
    Retrieves all works from a journal, given its ISSN, which match
    a set of keywords and are within a date range.

    Calling a CrossrefSearch object performs the search. A search object can be
    called with the arguments `display_progress_bar` (True/False; default=True) to
    toggle the display of a search progress bar, `select` (True/False; default=False), and
    `select_fields` (list) to query only a subset of metadata for each journal
    article.

    If select is False, a full (memory and time intensive) search is performed,
    fetching all metadata associated with each journal work.

    If select is True, a subset of fields to fetch can be specified using the
    select_fields parameter. Check the Crossref REST API doc
    for details on which field names are permissible.

    Note:
        Performing a search with no keywords and select=False can be very time- and memory- intensive.
        The search object will complain when such a search is performed.

    Args:
        ISSN (str): Journal (web) ISSN.
        type (str): Type of works to fetch (default="journal-article").
        keyword_list (list): List of keywords (str) to query with (default=None).
        from_date (str): Fetch articles published from (and after) this date (format="YYYY-MM-DD", default=None).
        until_date (str): Fetch articles published until this date (format="YYYY-MM-DD", default=None).
        batch_size (int): Number of works to fetch in each batch (default=20).
        sort_order (str): Order in which to sort works by date ("asc" or "desc", default="desc").

    Attributes:
        ISSN (str): Journal (web) ISSN.
        type (str): Type of works to fetch (default="journal-article").
        keyword_list (list): List of keywords (str) to query with.
        from_date (str): Fetch articles published from (and after) this date (format="YYYY-MM-DD").
        until_date (str): Fetch articles published until this date (format="YYYY-MM-DD").
        batch_size (int): Number of works to fetch in each batch (default=20).
        sort_order (str): Order in which to sort works by date ("asc" or "desc", default="desc").
        results (list): List of dictionaries, each dictionary corresponds to a work.

    Examples:
        >>> search = CrossrefSearch(ISSN="1520-5126", keyword_list=["hydration"], from_date="2018-01-01", until_date="2020-01-01")
        >>> search()
        >>> len(search)
        13
        >>> ds = search.get_DOIDataset()
        >>> ds.to_df()
                             DOI
        0   10.1021/jacs.9b09103
        1   10.1021/jacs.9b06862
        2   10.1021/jacs.9b09111
        3   10.1021/jacs.9b05874
        4   10.1021/jacs.9b02820
        5   10.1021/jacs.9b05136
        6   10.1021/jacs.9b02742
        7   10.1021/jacs.9b00577
        8   10.1021/jacs.8b11448
        9   10.1021/jacs.8b12877
        10  10.1021/jacs.8b11667
        11  10.1021/jacs.8b08298
        12  10.1021/jacs.7b11537
    """
    def __init__(self, ISSN="", type='journal-article', keyword_list=None, from_date=None,
                 until_date=None, batch_size=20, sort_order='desc'):
        self.ISSN = ISSN
        self.type = type
        self.keyword_list = keyword_list
        self.from_date = from_date
        self.until_date = until_date
        self.batch_size = batch_size
        self.sort_order = sort_order

        # Results
        self.results = []

    # Properties
    def __len__(self):
        return len(self.results)

    # Search workers (class methods)
    @classmethod
    def _check_issn_exists(cls, issn: str):
        """Checks if ISSN exists in Crossref.

        Args:
            issn (str): ISSN of journal

        Returns:
            bool
        """
        components = OrderedDict([("journals", str(issn))])
        query = CrossrefQuery(components)
        query()
        return query.response.status_code == 200

    @classmethod
    def _fetch_count(cls, issn: str, query_params=OrderedDict()):
        """Fetches number of works in journal that match criteria.

        Args:
            issn (str): ISSN of journal to check
            query_params (collections.OrderedDict): Parameters for query (default={})

        Returns:
            int

        Raises:
            SearchError if ISSN does not exist, or if unable to decode JSON response.
        """
        # Retrive summary of results only
        query_params['rows'] = 0

        if cls._check_issn_exists(issn):
            components = OrderedDict([("journals", str(issn)),
                                      ("works", None)])
            query = CrossrefQuery(components,
                                  query_params=query_params)
            query()

            try:
                data = query.response.json()
                return data['message']['total-results']

            except Exception:
                raise SearchError("Cannot decode results for the ISSN %s" % issn)
        else:
            raise SearchError("ISSN %s is not indexed in Crossref." % issn)

    @classmethod
    def _fetch_batch(cls, issn: str, query_params=OrderedDict(), size=20,
                     offset=0):
        """Fetches a batch of works.

        Args:
            issn (str): ISSN of journal to check
            query_params (collections.OrderedDict): Parameters for query (default={})
            size (int): Batch size (default=20)
            offset (int): Offset to fetch results from (default=0)

        Returns:
            data (dict): JSON response as Python dictionary.

        Raises:
            SearchError if ISSN does not exist, or if unable to decode JSON response.
        """
        # Set rows and offset
        query_params['rows'] = size
        query_params['offset'] = offset

        if cls._check_issn_exists(issn):
            components = OrderedDict([("journals", str(issn)),
                                      ("works", None)])
            query = CrossrefQuery(components,
                                  query_params=query_params)
            query()

            try:
                data = query.response.json()
                return data['message']

            except Exception:
                raise SearchError("Cannot decode results for offset %d of the ISSN %s" % (offset, issn))
        else:
            raise SearchError("ISSN %s is not indexed in Crossref." % issn)

    @classmethod
    def _extract_fields(cls, json_item, field_list, field_parsers_list):
        """Extracts data corresponding to given list of fields from the JSON
        response returned by the Crossref API.

        Args:
            json_item (dict): Item dictionary extracted from JSON response object.
            field_list (list): List of field names.
            field_parsers_list (list): List of field parser functions to parse field values.
                If field parser is None, the output string will be appended as is.

        Returns:
            output_item (list): List of extracted field values.

        Raises:
            SearchError if field parser is missing.
        """
        output_item = []
        for fidx in range(len(field_list)):
            field = field_list[fidx]

            try:
                field_parser = field_parsers_list[fidx]

            except IndexError:
                raise SearchError("No field parser corresponding to field %s was provided." % field)

            try:
                extracted_field = json_item[field]
                if field_parser is not None:
                    output_item.append(field_parser(extracted_field))
                else:
                    output_item.append(extracted_field)

            except KeyError:
                output_item.append("")

        return output_item

    def dry_run(self, select=False, select_fields=[]):
        """
        How many works will this search fetch?
        """
        query_params = OrderedDict()
        if self.keyword_list is None:
            if not select:
                warnings.warn("Search with no keywords and no select can be slow and memory intensive. Consider setting select=True and using select_fields to fetch only a subset of fields.")
        elif len(self.keyword_list) == 0:
            if not select:
                warnings.warn("Search with no keywords and no select can be slow and memory intensive. Consider setting select=True and using select_fields to fetch only a subset of fields.")
        else:
            query_params['query'] = "+".join(self.keyword_list)

        if self.from_date is not None and self.until_date is not None:
            query_params['filter'] = "from-pub-date:{},until-pub-date:{}".format(
                                     self.from_date, self.until_date)
        elif self.from_date is not None:
            query_params['filter'] = "from-pub-date:{}".format(
                                     self.from_date)
        elif self.until_date is not None:
            query_params['filter'] = "until-pub-date:{}".format(
                                     self.until_date)

        query_params['facet'] = "type-name:{}".format(self.type)
        query_params['sort'] = "published"
        query_params['order'] = self.sort_order

        if(select):
            if(len(select_fields) == 0):
                raise SearchError("select_fields cannot be empty when select is True.")
            query_params['select'] = ",".join(select_fields)

        total_items = self._fetch_count(self.ISSN, query_params)

        return total_items

    def __call__(self, display_progress_bar=True, select=False, select_fields=[]):
        query_params = OrderedDict()
        if self.keyword_list is None:
            if not select:
                warnings.warn("Search with no keywords and no select can be slow and memory intensive. Consider setting select=True and using select_fields to fetch only a subset of fields.")
        elif len(self.keyword_list) == 0:
            if not select:
                warnings.warn("Search with no keywords and no select can be slow and memory intensive. Consider setting select=True and using select_fields to fetch only a subset of fields.")
        else:
            query_params['query'] = "+".join(self.keyword_list)

        if self.from_date is not None and self.until_date is not None:
            query_params['filter'] = "from-pub-date:{},until-pub-date:{}".format(
                                     self.from_date, self.until_date)
        elif self.from_date is not None:
            query_params['filter'] = "from-pub-date:{}".format(
                                     self.from_date)
        elif self.until_date is not None:
            query_params['filter'] = "until-pub-date:{}".format(
                                     self.until_date)

        query_params['facet'] = "type-name:{}".format(self.type)
        query_params['sort'] = "published"
        query_params['order'] = self.sort_order

        if(select):
            if(len(select_fields) == 0):
                raise SearchError("select_fields cannot be empty when select is True.")
            query_params['select'] = ",".join(select_fields)

        total_items = self._fetch_count(self.ISSN, query_params)
        logger.info("Fetching {} works.".format(total_items))

        offsets = range(total_items)[::self.batch_size]

        if display_progress_bar:
            if GlobalConfig.streamlit:
                offsets = stqdm(offsets, desc="Fetching {} batches of {} articles".format(len(offsets), self.batch_size))
            else:
                offsets = tqdm(offsets, desc="Fetching {} batches of {} articles".format(len(offsets), self.batch_size))
        else:
            logger.info("Fetching {} batches of {} articles.".format(len(offsets), self.batch_size))

        for offset in offsets:
            try:
                batch = self._fetch_batch(self.ISSN, query_params, self.batch_size,
                                          offset)
                self.results += (batch['items'])

            except (SearchError, Exception):
                # Do not fail in the middle of a search
                warnings.warn("Error in fetching batch of items from %d - %d. Omitting these items from search results." % (offset, offset + self.batch_size))

    def get_DOIDataset(self):
        """
        Extracts DOIs from search results and returns them as a DOIDataset object.

        Returns:
            DOIDataset
        """
        DOIlist = []
        for work in self.results:
            try:
                DOIlist.append(work['DOI'])
            except Exception:
                warnings.warn("Work {} did not contain a DOI. Omitting from search results.".format(str(work)))
        logger.debug(DOIlist)
        return DOIDataset(DOIlist)

    def get_CitationsDataset(self, field_list=[], field_parsers_list=[]):
        """
        Parses a selection of fields from search results and returns them as a CitationsDataset object.

        Args:
            field_list (list): Names of fields to parse (see Crossref REST API doc for permissible field name values).
            field_parsers_list (list): List of field parser functions corresponding to each field name. A `None` value means that no parser
                is needed for that field.

        Returns:
            CitationsDataset

        Example:
            >>> search = handsearch.CrossrefSearch(ISSN="1520-5126", keyword_list=["hydration"], from_date="2018-01-01",
            ...                                    until_date="2020-01-01")
            >>> search(select=True, select_fields=['DOI', 'URL', 'title', 'author', 'issued'])
            >>> ds = search.get_CitationsDataset(field_list=['DOI', 'URL', 'title', 'author', 'issued'],
            ...                                  field_parsers_list=[None, None, parsers.crossref_title_parser,
            ...                                                      parsers.crossref_authors_parser, parsers.crossref_date_parser])
        """
        Citationlist = []
        for work in self.results:
            try:
                fields = self._extract_fields(work, field_list, field_parsers_list)
                Citationlist.append(fields)
            except Exception:
                warnings.warn("Could not extract data from work {}. Omitting from search results.".format(str(work)))
        logger.debug(Citationlist)
        return CitationsDataset(field_list, Citationlist)

    def get_RISDataset(self, extra_field_list=[], extra_field_parser_list=[], extra_field_rispy_tags=[]):
        """
        Extracts DOIs from search results and fetches RIS data for each DOI using
        Crossref's content negotiation service.

        Extra fields in the search results that are not automatically populated
        by Crossref's content negotation service can be mapped to the RIS format (through rispy's mapping)
        using the `extra_fields`, `extra_field_parser_list`, and `extra_field_rispy_tags`
        arguments.

        Args:
            extra_field_list (list): List of extra fields to parse and include in RIS file (see Crossref REST API doc for permissible field name values).
            extra_field_parser_list (list): List of field parser functions corresponding to each extra field name. A `None` value means that no parser
                is needed for that field.
            extra_field_rispy_tags (list): List of rispy tags for each extra field.
        """
        RIS_dicts = []

        if GlobalConfig.streamlit:
            results = stqdm(self.results, desc="Converting results to RIS format.")
        else:
            results = tqdm(self.results, desc="Converting results to RIS format.")

        for work in results:
            try:
                # Extract DOI and extra fields
                doi_plus = self._extract_fields(work, ['DOI'] + extra_field_list, [None] + extra_field_parser_list)
                ris_ref = crossref_negotiate_ris(doi_plus[0])[0]

            except Exception:
                try:
                    doi = work['DOI']
                    ris_ref = {'type_of_reference': 'JOUR', 'doi': doi}
                    warnings.warn("Failed to get RIS metadata for DOI %s. Appending just the DOI to the RIS dataset." % doi)

                except Exception:
                    warnings.warn("Failed to get DOI from work {}. Skipping.".format(str(work)))
                    continue

            # Add in extra fields
            for fieldidx, field in enumerate(extra_field_list):
                tag = extra_field_rispy_tags[fieldidx]
                ris_ref[tag] = doi_plus[1 + fieldidx]

            # Add to list
            RIS_dicts.append(ris_ref)

        return RISDataset(RIS_dicts)
